@@ -12,6 +12,10 @@ class RayParams:
     """A class used to store the parameters used by Ray.
 
     Attributes:
+        external_addresses (str): The address of external Redis server to
+            connect to, in format of "ip1:port1,ip2:port2,...".  If this
+            address is provided, then ray won't start Redis instances in the
+            head node but use external Redis server(s) instead.
         redis_address (str): The address of the Redis server to connect to. If
             this address is not provided, then this command will start Redis, a
             raylet, a plasma store, a plasma manager, and some workers.
@@ -72,6 +76,8 @@ class RayParams:
             worker_setup_hook to set up the environment for the worker process.
         worker_setup_hook (str): The module path to a Python function that will
             be imported and run to set up the environment for the worker.
+        runtime_env_setup_hook (str): The module path to a Python function that
+            will be imported and run to set up the runtime env in agent.
         huge_pages: Boolean flag indicating whether to start the Object
             Store with hugetlbfs support. Requires plasma_directory.
         include_dashboard: Boolean flag indicating whether to start the web
@@ -84,6 +90,8 @@ class RayParams:
             external machines.
         dashboard_port: The port to bind the dashboard server to.
             Defaults to 8265.
+        dashboard_agent_listen_port: The port for dashboard agents to listen on
+            for HTTP requests.
         logging_level: Logging level, default will be logging.INFO.
         logging_format: Logging format, default contains a timestamp,
             filename, line number, and message. See ray_constants.py.
@@ -109,9 +117,13 @@ class RayParams:
             failure.
         start_initial_python_workers_for_first_job (bool): If true, start
             initial Python workers for the first job on the node.
+        ray_debugger_external (bool): If true, make the Ray debugger for a
+            worker available externally to the node it is running on. This will
+            bind on 0.0.0.0 instead of localhost.
     """
 
     def __init__(self,
+                 external_addresses=None,
                  redis_address=None,
                  num_cpus=None,
                  num_gpus=None,
@@ -141,10 +153,13 @@ class RayParams:
                  worker_path=None,
                  setup_worker_path=None,
                  worker_setup_hook=ray_constants.DEFAULT_WORKER_SETUP_HOOK,
+                 runtime_env_setup_hook=ray_constants.
+                 DEFAULT_RUNTIME_ENV_SETUP_HOOK,
                  huge_pages=False,
                  include_dashboard=None,
                  dashboard_host=ray_constants.DEFAULT_DASHBOARD_IP,
                  dashboard_port=ray_constants.DEFAULT_DASHBOARD_PORT,
+                 dashboard_agent_listen_port=0,
                  logging_level=logging.INFO,
                  logging_format=ray_constants.LOGGER_FORMAT,
                  plasma_store_socket_name=None,
@@ -153,6 +168,7 @@ class RayParams:
                  include_log_monitor=None,
                  autoscaling_config=None,
                  start_initial_python_workers_for_first_job=False,
+                 ray_debugger_external=False,
                  _system_config=None,
                  enable_object_reconstruction=False,
                  metrics_agent_port=None,
@@ -161,6 +177,7 @@ class RayParams:
                  no_monitor=False,
                  lru_evict=False):
         self.object_ref_seed = object_ref_seed
+        self.external_addresses = external_addresses
         self.redis_address = redis_address
         self.num_cpus = num_cpus
         self.num_gpus = num_gpus
@@ -189,10 +206,12 @@ class RayParams:
         self.worker_path = worker_path
         self.setup_worker_path = setup_worker_path
         self.worker_setup_hook = worker_setup_hook
+        self.runtime_env_setup_hook = runtime_env_setup_hook
         self.huge_pages = huge_pages
         self.include_dashboard = include_dashboard
         self.dashboard_host = dashboard_host
         self.dashboard_port = dashboard_port
+        self.dashboard_agent_listen_port = dashboard_agent_listen_port
         self.plasma_store_socket_name = plasma_store_socket_name
         self.raylet_socket_name = raylet_socket_name
         self.temp_dir = temp_dir
@@ -204,6 +223,7 @@ class RayParams:
         self.no_monitor = no_monitor
         self.start_initial_python_workers_for_first_job = (
             start_initial_python_workers_for_first_job)
+        self.ray_debugger_external = ray_debugger_external
         self._system_config = _system_config or {}
         self._enable_object_reconstruction = enable_object_reconstruction
         self._check_usage()
