@@ -10,11 +10,6 @@ import json
 import time
 import traceback
 
-import aiohttp
-import aiohttp.web
-import aiohttp_cors
-import psutil
-from aiohttp import hdrs
 from grpc.experimental import aio as aiogrpc
 
 import ray
@@ -27,6 +22,14 @@ from ray.core.generated import agent_manager_pb2
 from ray.core.generated import agent_manager_pb2_grpc
 from ray._private.ray_logging import setup_component_logger
 
+# All third-party dependencies that are not included in the minimal Ray
+# installation must be included in this file. This allows us to determine if
+# the agent has the necessary dependencies to be started.
+from ray.new_dashboard.optional_deps import aiohttp, aiohttp_cors, hdrs
+
+# Import psutil after ray so the packaged version is used.
+import psutil
+
 try:
     create_task = asyncio.create_task
 except AttributeError:
@@ -51,8 +54,10 @@ class DashboardAgent(object):
                  log_dir=None,
                  metrics_export_port=None,
                  node_manager_port=None,
+                 listen_port=0,
                  object_store_name=None,
-                 raylet_name=None):
+                 raylet_name=None,
+                 logging_params=None):
         """Initialize the DashboardAgent object."""
         # Public attributes are accessible for all agent modules.
         self.ip = node_ip_address
@@ -66,8 +71,10 @@ class DashboardAgent(object):
         self.dashboard_agent_port = dashboard_agent_port
         self.metrics_export_port = metrics_export_port
         self.node_manager_port = node_manager_port
+        self.listen_port = listen_port
         self.object_store_name = object_store_name
         self.raylet_name = raylet_name
+        self.logging_params = logging_params
         self.node_id = os.environ["RAY_NODE_ID"]
         # TODO(edoakes): RAY_RAYLET_PID isn't properly set on Windows. This is
         # only used for fate-sharing with the raylet and we need a different
@@ -163,7 +170,7 @@ class DashboardAgent(object):
 
         runner = aiohttp.web.AppRunner(app)
         await runner.setup()
-        site = aiohttp.web.TCPSite(runner, self.ip, 0)
+        site = aiohttp.web.TCPSite(runner, "0.0.0.0", self.listen_port)
         await site.start()
         http_host, http_port, *_ = site._server.sockets[0].getsockname()
         logger.info("Dashboard agent http address: %s:%s", http_host,
@@ -235,6 +242,12 @@ if __name__ == "__main__":
         type=str,
         default=None,
         help="The socket name of the plasma store")
+    parser.add_argument(
+        "--listen-port",
+        required=False,
+        type=int,
+        default=0,
+        help="Port for HTTP server to listen on")
     parser.add_argument(
         "--raylet-name",
         required=True,
@@ -317,13 +330,14 @@ if __name__ == "__main__":
 
     args = parser.parse_args()
     try:
-        setup_component_logger(
+        logging_params = dict(
             logging_level=args.logging_level,
             logging_format=args.logging_format,
             log_dir=args.log_dir,
             filename=args.logging_filename,
             max_bytes=args.logging_rotate_bytes,
             backup_count=args.logging_rotate_backup_count)
+        setup_component_logger(**logging_params)
 
         # The dashboard is currently broken on Windows.
         # https://github.com/ray-project/ray/issues/14026.
@@ -347,8 +361,10 @@ if __name__ == "__main__":
             log_dir=args.log_dir,
             metrics_export_port=args.metrics_export_port,
             node_manager_port=args.node_manager_port,
+            listen_port=args.listen_port,
             object_store_name=args.object_store_name,
-            raylet_name=args.raylet_name)
+            raylet_name=args.raylet_name,
+            logging_params=logging_params)
 
         loop = asyncio.get_event_loop()
         loop.run_until_complete(agent.run())
