@@ -9,7 +9,7 @@ import numpy as np
 import time
 
 from ray.tune.callback import Callback
-from ray.tune.logger import pretty_print
+from ray.tune.logger import pretty_print, logger
 from ray.tune.result import (DEFAULT_METRIC, EPISODE_REWARD_MEAN,
                              MEAN_ACCURACY, MEAN_LOSS, TRAINING_ITERATION,
                              TIME_TOTAL_S, TIMESTEPS_TOTAL, AUTO_RESULT_KEYS)
@@ -29,6 +29,12 @@ except ImportError:
                       "Please re-run 'pip install ray[tune]' or "
                       "'pip install ray[rllib]'.")
 
+try:
+    class_name = get_ipython().__class__.__name__
+    IS_NOTEBOOK = True if "Terminal" not in class_name else False
+except NameError:
+    IS_NOTEBOOK = False
+
 
 class ProgressReporter:
     """Abstract class for experiment progress reporting.
@@ -57,6 +63,13 @@ class ProgressReporter:
         """
         raise NotImplementedError
 
+    def set_search_properties(self, metric: Optional[str],
+                              mode: Optional[str]):
+        return True
+
+    def set_total_samples(self, total_samples: int):
+        pass
+
 
 class TuneReporterBase(ProgressReporter):
     """Abstract base class for the default Tune reporters.
@@ -94,6 +107,8 @@ class TuneReporterBase(ProgressReporter):
         metric (str): Metric used to determine best current trial.
         mode (str): One of [min, max]. Determines whether objective is
             minimizing or maximizing the metric attribute.
+        sort_by_metric (bool): Sort terminated trials by metric in the
+            intermediate table. Defaults to False.
     """
 
     # Truncated representations of column names (to accommodate small screens).
@@ -121,7 +136,8 @@ class TuneReporterBase(ProgressReporter):
             infer_limit: int = 3,
             print_intermediate_tables: Optional[bool] = None,
             metric: Optional[str] = None,
-            mode: Optional[str] = None):
+            mode: Optional[str] = None,
+            sort_by_metric: bool = False):
         self._total_samples = total_samples
         self._metrics_override = metric_columns is not None
         self._inferred_metrics = {}
@@ -143,6 +159,11 @@ class TuneReporterBase(ProgressReporter):
         self._metric = metric
         self._mode = mode
 
+        if metric is None or mode is None:
+            self._sort_by_metric = False
+        else:
+            self._sort_by_metric = sort_by_metric
+
     def set_search_properties(self, metric: Optional[str],
                               mode: Optional[str]):
         if self._metric and metric:
@@ -268,7 +289,10 @@ class TuneReporterBase(ProgressReporter):
                     force_table=self._print_intermediate_tables,
                     fmt=fmt,
                     max_rows=max_progress,
-                    done=done))
+                    done=done,
+                    metric=self._metric,
+                    mode=self._mode,
+                    sort_by_metric=self._sort_by_metric))
             messages.append(
                 trial_errors_str(trials, fmt=fmt, max_rows=max_error))
 
@@ -354,6 +378,8 @@ class JupyterNotebookReporter(TuneReporterBase):
         metric (str): Metric used to determine best current trial.
         mode (str): One of [min, max]. Determines whether objective is
             minimizing or maximizing the metric attribute.
+        sort_by_metric (bool): Sort terminated trials by metric in the
+            intermediate table. Defaults to False.
     """
 
     def __init__(
@@ -368,11 +394,23 @@ class JupyterNotebookReporter(TuneReporterBase):
             infer_limit: int = 3,
             print_intermediate_tables: Optional[bool] = None,
             metric: Optional[str] = None,
-            mode: Optional[str] = None):
+            mode: Optional[str] = None,
+            sort_by_metric: bool = False):
         super(JupyterNotebookReporter, self).__init__(
             metric_columns, parameter_columns, total_samples,
             max_progress_rows, max_error_rows, max_report_frequency,
-            infer_limit, print_intermediate_tables, metric, mode)
+            infer_limit, print_intermediate_tables, metric, mode,
+            sort_by_metric)
+
+        if not IS_NOTEBOOK:
+            logger.warning(
+                "You are using the `JupyterNotebookReporter`, but not "
+                "IPython/Jupyter-compatible environment was detected. "
+                "If this leads to unformatted output (e.g. like "
+                "<IPython.core.display.HTML object>), consider passing "
+                "a `CLIReporter` as the `progress_reporter` argument "
+                "to `tune.run()` instead.")
+
         self._overwrite = overwrite
 
     def report(self, trials: List[Trial], done: bool, *sys_info: Dict):
@@ -417,6 +455,8 @@ class CLIReporter(TuneReporterBase):
         metric (str): Metric used to determine best current trial.
         mode (str): One of [min, max]. Determines whether objective is
             minimizing or maximizing the metric attribute.
+        sort_by_metric (bool): Sort terminated trials by metric in the
+            intermediate table. Defaults to False.
     """
 
     def __init__(
@@ -430,12 +470,14 @@ class CLIReporter(TuneReporterBase):
             infer_limit: int = 3,
             print_intermediate_tables: Optional[bool] = None,
             metric: Optional[str] = None,
-            mode: Optional[str] = None):
+            mode: Optional[str] = None,
+            sort_by_metric: bool = False):
 
         super(CLIReporter, self).__init__(
             metric_columns, parameter_columns, total_samples,
             max_progress_rows, max_error_rows, max_report_frequency,
-            infer_limit, print_intermediate_tables, metric, mode)
+            infer_limit, print_intermediate_tables, metric, mode,
+            sort_by_metric)
 
     def report(self, trials: List[Trial], done: bool, *sys_info: Dict):
         print(self._progress_str(trials, done, *sys_info))
@@ -460,7 +502,7 @@ def memory_debug_str():
             round(used_gb, 1), round(total_gb, 1), warn)
     except ImportError:
         return ("Unknown memory usage. Please run `pip install psutil` "
-                "(or ray[debug]) to resolve)")
+                "to resolve)")
 
 
 def _get_trials_by_state(trials: List[Trial]):
@@ -478,7 +520,10 @@ def trial_progress_str(
         force_table: bool = False,
         fmt: str = "psql",
         max_rows: Optional[int] = None,
-        done: bool = False):
+        done: bool = False,
+        metric: Optional[str] = None,
+        mode: Optional[str] = None,
+        sort_by_metric: bool = False):
     """Returns a human readable message for printing to the console.
 
     This contains a table where each row represents a trial, its parameters
@@ -503,6 +548,11 @@ def trial_progress_str(
         max_rows (int): Maximum number of rows in the trial table. Defaults to
             unlimited.
         done (bool): True indicates that the tuning run finished.
+        metric (str): Metric used to sort trials.
+        mode (str): One of [min, max]. Determines whether objective is
+            minimizing or maximizing the metric attribute.
+        sort_by_metric (bool): Sort terminated trials by metric in the
+            intermediate table. Defaults to False.
     """
     messages = []
     delim = "<br>" if fmt == "html" else "\n"
@@ -529,7 +579,8 @@ def trial_progress_str(
 
     if force_table or (has_verbosity(Verbosity.V2_TRIAL_NORM) and done):
         messages += trial_progress_table(trials, metric_columns,
-                                         parameter_columns, fmt, max_rows)
+                                         parameter_columns, fmt, max_rows,
+                                         metric, mode, sort_by_metric)
 
     return delim.join(messages)
 
@@ -539,20 +590,30 @@ def trial_progress_table(
         metric_columns: Union[List[str], Dict[str, str]],
         parameter_columns: Union[None, List[str], Dict[str, str]] = None,
         fmt: str = "psql",
-        max_rows: Optional[int] = None):
+        max_rows: Optional[int] = None,
+        metric: Optional[str] = None,
+        mode: Optional[str] = None,
+        sort_by_metric: bool = False):
     messages = []
     num_trials = len(trials)
     trials_by_state = _get_trials_by_state(trials)
 
+    # Sort terminated trials by metric and mode, descending if mode is "max"
+    if sort_by_metric:
+        trials_by_state[Trial.TERMINATED] = sorted(
+            trials_by_state[Trial.TERMINATED],
+            reverse=(mode == "max"),
+            key=lambda t: t.last_result[metric])
+
     state_tbl_order = [
         Trial.RUNNING, Trial.PAUSED, Trial.PENDING, Trial.TERMINATED,
         Trial.ERROR
     ]
-
     max_rows = max_rows or float("inf")
     if num_trials > max_rows:
         # TODO(ujvl): suggestion for users to view more rows.
-        trials_by_state_trunc = _fair_filter_trials(trials_by_state, max_rows)
+        trials_by_state_trunc = _fair_filter_trials(trials_by_state, max_rows,
+                                                    sort_by_metric)
         trials = []
         overflow_strs = []
         for state in state_tbl_order:
@@ -580,9 +641,11 @@ def trial_progress_table(
         metric_keys = list(metric_columns.keys())
     else:
         metric_keys = metric_columns
+
     metric_keys = [
         k for k in metric_keys if any(
-            t.last_result.get(k) is not None for t in trials)
+            unflattened_lookup(k, t.last_result, default=None) is not None
+            for t in trials)
     ]
 
     if not parameter_columns:
@@ -666,7 +729,8 @@ def best_trial_str(
 
 
 def _fair_filter_trials(trials_by_state: Dict[str, List[Trial]],
-                        max_trials: int):
+                        max_trials: int,
+                        sort_by_metric: bool = False):
     """Filters trials such that each state is represented fairly.
 
     The oldest trials are truncated if necessary.
@@ -687,12 +751,16 @@ def _fair_filter_trials(trials_by_state: Dict[str, List[Trial]],
                 no_change = False
                 max_trials -= 1
                 num_trials_by_state[state] += 1
-    # Sort by start time, descending.
-    sorted_trials_by_state = {
-        state: sorted(
-            trials_by_state[state], reverse=False, key=lambda t: t.trial_id)
-        for state in sorted(trials_by_state)
-    }
+    # Sort by start time, descending if the trails is not sorted by metric.
+    sorted_trials_by_state = dict()
+    for state in sorted(trials_by_state):
+        if state == Trial.TERMINATED and sort_by_metric:
+            sorted_trials_by_state[state] = trials_by_state[state]
+        else:
+            sorted_trials_by_state[state] = sorted(
+                trials_by_state[state],
+                reverse=False,
+                key=lambda t: t.trial_id)
     # Truncate oldest trials.
     filtered_trials = {
         state: sorted_trials_by_state[state][:num_trials_by_state[state]]
@@ -827,3 +895,20 @@ class TrialProgressCallback(Callback):
         print_result_str = ",".join(
             [f"{k}={v}" for k, v in print_result.items()])
         return print_result_str
+
+
+def detect_reporter(**kwargs) -> TuneReporterBase:
+    """Detect progress reporter class.
+
+    Will return a :class:`JupyterNotebookReporter` if a IPython/Jupyter-like
+    session was detected, and a :class:`CLIReporter` otherwise.
+
+    Keyword arguments are passed on to the reporter class.
+    """
+    if IS_NOTEBOOK:
+        kwargs.setdefault("overwrite",
+                          not has_verbosity(Verbosity.V2_TRIAL_NORM))
+        progress_reporter = JupyterNotebookReporter(**kwargs)
+    else:
+        progress_reporter = CLIReporter(**kwargs)
+    return progress_reporter
